--- Provision module for SIP pa2 video intercom & Paging device
-- 
-- @file	wildixpa.lua 
-- @author	Gennadiy Ganchev
-- @created 2017/10/03

module( ... or "wildixpa", package.seeall)

hints = {
	--[[
	======================
	Wildix SIP pa2
	----------------------
		200 OK.
		Server: Rapid Logic/1.1.
		MIME-version: 1.0.
		Date: Thu Jan  1 00:16:04 1970 GMT.
		Content-Type: text/html.
		Transfer-Encoding: chunked.
		Set-Cookie: auth=0a6403b3000003c4; path=/.
	--]]
	{
		 ["server"] = 'Rapid Logic/.*'
		,["status"] = "200"
	}
}

--[[
Parameters in config file should be left aligned with spaces (Fanvil's requirement)
<MCAST CFG MODULE>
Priority           :0
Enbale Priority    :0
--Mcast Addr--     :
MCAST1 Host        :239.1.1.10
MCAST1 Port        :10000
MCAST2 Host        :239.1.1.10
MCAST2 Port        :10002
MCAST2 Name        :pg3
MCAST1 Name        :pg1
]]--
local paramMaxLen = 19

local function __getTimeZoneLocation(timeZone)
	local location = nil
	local zone	 = nil
	local continent, location = string.match(timeZone or "", "^([%w%-%+]+)/*(.*)$")
	local t = {
			 ["Africa"]	= {}
			,["America"]   = {}
			,["Asia"]	  = {}
			,["Atlantic"]  = {}
			,["Australia"] = {}
			,["Canada"]	= {}
			,["Etc"]	   = {
				 ["GMT-14"] = {zone = "56", location = "0"}
				,["GMT-13"] = {zone = "52", location = "0"}
				,["GMT-12"] = {zone = "48", location = "0"}
				,["GMT-11"] = {zone = "44", location = "0"}
				,["GMT-10"] = {zone = "40", location = "0"}
				,["GMT-9"]  = {zone = "36", location = "0"}
				,["GMT-8"]  = {zone = "32", location = "0"}
				,["GMT-7"]  = {zone = "28", location = "0"}
				,["GMT-6"]  = {zone = "24", location = "0"}
				,["GMT-5"]  = {zone = "20", location = "0"}
				,["GMT-4"]  = {zone = "16", location = "0"}
				,["GMT-3"]  = {zone = "12", location = "0"}
				,["GMT-2"]  = {zone = "8", location = "0"}
				,["GMT-1"]  = {zone = "4", location = "0"}
				,["GMT-0"]  = {zone = "4", location = "0"}
				,["GMT"]	= {zone = "4", location = "0"}
				,["GMT0"]   = {zone = "4", location = "0"}
				,["GMT+0"]  = {zone = "4", location = "0"}
				,["GMT+1"]  = {zone = "-4", location = "0"}
				,["GMT+2"]  = {zone = "-8", location = "0"}
				,["GMT+3"]  = {zone = "-12", location = "0"}
				,["GMT+4"]  = {zone = "-16", location = "0"}
				,["GMT+5"]  = {zone = "-20", location = "0"}
				,["GMT+6"]  = {zone = "-24", location = "0"}
				,["GMT+7"]  = {zone = "-28", location = "0"}
				,["GMT+8"]  = {zone = "-32", location = "0"}
				,["GMT+9"]  = {zone = "-36", location = "0"}
				,["GMT+10"] = {zone = "-40", location = "0"}
				,["GMT+11"] = {zone = "-44", location = "0"}
			}
			,["Europe"]	= {
				 ["Berlin"]	 = {zone = "4", location = "11"} -- GERMANY
				,["Paris"]	  = {zone = "4", location = "10"} -- FRANCE
				,["Rome"]	   = {zone = "4", location = "13"} -- ITALY
				,["Amsterdam"]  = {zone = "4", location = "16"} -- HOLLAND
				,["Copenhagen"] = {zone = "4", location = "9"}  -- DENMARK
				,["Kiev"]	   = {zone = "8", location = "14"} -- UKRAINE
				,["Brussels"]   = {zone = "4", location = "3"}  -- BELGIUM
				,["Vienna"]	 = {zone = "4", location = "2"}  -- AUSTRIA
				,["London"]	 = {zone = "0", location = "7"}  -- ENGLAND
			}
			,["Indian"]	= {}
			,["Pacific"]   = {}
			,["SystemV"]   = {}
			,["US"]		= {
				 ["Alaska"]		 = {zone = "-36", location = "1"}
				,["Aleutian"]	   = {zone = "-44", location = "0"}
				,["Arizona"]		= {zone = "-28", location = "0"}
				,["Central"]		= {zone = "-24", location = "4"}
				,["Eastern"]		= {zone = "-20", location = "4"}
				,["East-Indiana"]   = {zone = "-20", location = "4"}
				,["Hawaii"]		 = {zone = "-40", location = "1"}
				,["Indiana-Starke"] = {zone = "-24", location = "4"}
				,["Michigan"]	   = {zone = "-20", location = "0"}
				,["Mountain"]	   = {zone = "-28", location = "0"}
				,["Pacific"]		= {zone = "-32", location = "0"}
				,["Samoa"]		  = {zone = "-44", location = "1"}
			}
	}
	
	if t[continent] and t[continent][location] then
		zone = t[continent][location]["zone"]
		location  = t[continent][location]["location"]
	end
	
	return zone, location
end

--- Verifying device HTML request and retreiving information
-- @param host		A verifying host data
-- @param port		Not used data
-- @return table with verifying device information
function verify (host, port)
	local logger     = logger
	local http       = require"core.httpClient":new()
	local xml        = require"core.htmlparser"
	local helper     = require"common.helper"
	local connection = require"common.core".connection
	
	local is_ssl_negotiate = 443 == port.number
	local PORT_NUMBER      = is_ssl_negotiate and not gconf.HTTPForceUse and 443 or 80
	local t                = {["ssl"] = is_ssl_negotiate, ["model"] = ""}
	
	local gpassword  = helper.get_gpassword()
	local auth_table = {"admin:admin", "admin:" .. gpassword}

	local sql          = nil
	local cursor       = nil
	local token_pass   = nil
	local dublicate_ip = nil
	if "" == host.mac_addr_str then
		sql    = [[ SELECT token, count(id) FROM autoprov WHERE param=%q AND value LIKE '%%"address":%%"%s"%%' AND value LIKE '%%"model":%%"w-pa"%%' ]]
		cursor = connection:execute(sql, "device", host.ip or "")
		token_pass, dublicate_ip = cursor:fetch(); cursor:close()
		if 1 < tonumber(dublicate_ip) then
			logger:error("Found %s duplicated address %q at provisioning database", dublicate_ip, host.ip or "")
			return t
		end
	else
		sql 	   = "SELECT token FROM autoprov WHERE param=%q AND mac=%q"
		cursor 	   = connection:execute(sql, "device", host.mac_addr_str)
		token_pass = cursor:fetch(); cursor:close()
	end

	t["token_type"] = helper.is_new_token(token_pass)
	if "new" == t["token_type"] then
		table.insert(auth_table, 1, "admin:"..token_pass)
		t["token"] = token_pass
	end
	logger:debug(auth_table, "auth_table")

	--- authorization procedure
	local answer = http:rawget(host.ip, PORT_NUMBER, "/") or {}
	local cookie = answer.headers and answer.headers["set-cookie"]
	      cookie = cookie and cookie:match("^(auth=%w+);")
	local nonce  = cookie and cookie:match("^auth=(%w+)")
	
	if nonce then
		for _, entry in pairs(auth_table) do
			local login, passwd = entry:match("(.*):(.*)")
			local encodedPasswd = require"md5".digest(login..":"..passwd..":"..nonce)
			local data          = { encoded = login..":"..encodedPasswd, ReturnPage = "/" }
			answer        = http:rawpost(host.ip, PORT_NUMBER, "/", data, { headers = {cookie = cookie}})
			if answer and answer.body and not answer.body:find("User Name or Password Error!") then
				t["auth"]   = {user = login, pass = passwd}
				t["cookie"] = cookie
				break
			end
		end
	end
	if not t["auth"] then
		logger:error("Wrong authorization for host '%s'", host.ip)
		return t
	end
	
	--- collect device data
	local answer = http:rawget(host.ip, PORT_NUMBER, "/information.htm", {headers = {cookie = cookie}})
	if not answer or 200 ~= answer.code or not answer.body then
		logger:error("Could not retrieve device information for host '%s'", host.ip)
		return t
	end
	local doc = xml:new(answer.body)

	--WMS-6021
	local function __remove_tag(txt)
		return string.match(txt or "", ".+>(.-)<.+")
	end

	if "" == host.mac_addr_str then
		host.mac_addr_str = __remove_tag(doc:find('//div[@class="content"]/div[@class="navbar"]/div[4]/table/tr[2]/td[2]'))
	end
	if host.mac_addr_str then
		t["mac_addr_str"] = host.mac_addr_str:gsub(":", ""):lower()
	end

	local model = __remove_tag(doc:find('//div[@class="content"]/div[@class="navbar"]/div[2]/table/tr[1]/td[2]'))
	if not model or "" == model then
		logger:error("Not found model for host '%s'", host.ip)
		return t
	end

	t["model"] = model:lower()

	local fwVer    = __remove_tag(doc:find('//div[@class="content"]/div[@class="navbar"]/div[2]/table/tr[3]/td[2]'))
	t["fwversion"] = fwVer

	return t
end

--- Provision action to updating provision url on device and executing autoprovision
-- @param host	A provision device data
-- @param port	Not used 
function provision (host, port, ...)
    local logger = logger
	local h      = require"common.helper"
	local http   = require"core.httpClient":new()
	
	local PORT_NUMBER = host.ssl and not gconf.HTTPForceUse and 443 or 80
	local prov_port   = host.ssl and not gconf.HTTPForceUse and "443" or "80"
	local prov_proto  = "80" == prov_port and "4" or "5"                        -- "4": "http", "5": "https"
	
	-- processing registrar address
	local registrar = nil
	local case      = "hosts"
	local cfg       = {
			 FailoverByDns	  = gconf.FailoverByDns
			,ForceResolveName = gconf.ForceResolveName
			,ForceResolveIp   = gconf.ForceResolveIp
		}
	local addr_t, err_msg = h.resolve_by_hosts(host.ip_src, nil, cfg)
	if not addr_t then
		if err_msg then	-- printing resolving error
			logger:warning("Hosts file was not processed: %s", err_msg)
		end
	else
		registrar = addr_t.domain
	end
	if not registrar then
		registrar, case, err_msg = h.get_registrar(host.ip_src, host.ip) 
	end
	logger:debug("Registrar address %q got from %s%s", registrar, case, err_msg and ". Errors: "..err_msg or "")

	-- Get security password
	local gpassword = nil
	if host.token_type == "new" then
		gpassword = host.token or h.get_gpassword()
	else
		gpassword = h.get_gpassword()
		if host.token_type == "old" then
			logger:info("Token has old format")
		end
	end

	-- Set options for next http requests
	local options = {
			 headers = { ['Cookie'] = host.cookie }
			,timeout = 2
		}

	-- Set parameters for autoprovision
	local data = {
			 FTP_AutoUser_R             = "wmsdevice"
			,FTP_AutoPasswd_R           = gpassword
			,FTP_AutoConfigEncKey_RW    = ""
			,FTP_ApCfgEncComKey_RW      = ""
			,SYS_SaveApInfo_RW          = "ON"
			,SYS_OptionCustom_R         = 66
			,SYS_PnPServer_R            = "224.0.1.75"
			,SYS_PnPPort_R              = 5060
			,SYS_PnPTransport_R         = 0
			,SYS_PnPInterval_R          = 1
			,FTP_AutoServer_R           = registrar..":"..prov_port
			,FTP_AutoFileName_R         = "?action=setaps"
			,FTP_ProtocolType           = prov_proto
			,FTP_AutoUpdateInterval     = 1
			,FTP_AutoDownloadMode       = 1  -- Update after reboot
			,SYS_TR069ServerType_RW     = 1
			,SYS_ACSServerURL_RW        = "0.0.0.0"
			,SYS_ACSUsername_RW         = "admin"
			,SYS_ACSPassword_RW         = gpassword
			,SYS_TR069PeriodInterval_RW = 3600
			,ReturnPage                 = "/provision.htm"
			,CheckBoxManager            = "SYS_SaveApInfo_RW,SYS_EnablePnP_RW,SYS_TR069Enable_RW,SYS_TR069AutoLogin_RW"
			,DefaultSubmit              = "Apply"
		}

	local answer = http:rawpost(host.ip, PORT_NUMBER, "/provision.htm", data, options) or {}
	logger:debug("Setting autoprovisioning parameters for %s - %s", host.model, answer.code or 408)

	answer = http:rawpost(host.ip, PORT_NUMBER, "/tools.htm", { DefaultReboot = "Reboot" }, options) or {}
	logger:info("Sending http request to %s for autoprovision now - %s", host.model, answer.code or 408)
end

--- Building and sending provision configuration
-- @param agent		A table with general device information
-- @param mixed		Some parameters. Currently not used		
function configuration (agent, request)
	if "table" ~= type(agent) or "table" ~= type(request) then
		return false
	end

	local h          = require"common.helper"
	local core       = require"common.core"
	local logger     = logger
	local connection = core.connection
	local json       = require"dkjson"
	local md5        = require"md5"
	local wmsldap    = require"core.wmsldap"
	local devFwInfo  = h.parse_firmware(agent["fw:version"])

	-- initial variables
	local cfg_t       = nil -- configuration table
	local prov_proto  = nil
	local prov_port   = nil

	-- load default values from resource directory
	local resource = core.load_resource("wildixpa")
	cfg_t          = resource.cfg_def or {}

	-- retreive registrar URL parameters
	local registrar    = agent["registrar"]
	local registrar_ip = false
	local cfg          = {
			 FailoverByDns    = gconf.FailoverByDns
			,ForceResolveName = gconf.ForceResolveName
			,ForceResolveIp   = gconf.ForceResolveIp
		}
	local addr_t, err_msg = h.resolve_by_hosts(registrar, nil, cfg)
	if err_msg then
		logger:warning("Hosts file processing error: %s", err_msg)
	end
	if addr_t then
		logger:debug(addr_t, "hosts_registrar")
		if addr_t.domain then
			registrar = addr_t.domain
		end
		if addr_t.ip then
			registrar_ip = addr_t.ip
		end
	end
	
	-- retrieve device information
	local token_pass = core.proceedDeviceInfo(agent).token

	local prov_proto = agent["ssl"] == "on" and not gconf.HTTPForceUse and "https" or "http"
	local prov_port  = gconf.HTTPForceUse and "80" or agent["alternate_port"] or agent["registrar:port"]

-----------------------------------------------------
--- Provision settings
-----------------------------------------------------	
	cfg_t["<AUTOUPDATE CONFIG MODULE>"]["Download Password"] = token_pass or h.get_gpassword()
	cfg_t["<AUTOUPDATE CONFIG MODULE>"]["Download Server IP"] = registrar..":"..prov_port
	cfg_t["<AUTOUPDATE CONFIG MODULE>"]["Download Protocol"] = "https" == prov_proto and "5" or "4"

	-- retreiving information about device firmware
	local fw_cfg = nil
	local firmwares = core.get_firmware_pack({["vendor"] = "wildixpa"})
	if firmwares then
		fw_cfg = firmwares[agent["model"]]
	end
	if not fw_cfg then
		logger:error("Firmware pack not found for device %q", agent["model"])
		return false
	end
	if fw_cfg and fw_cfg["version"] ~= agent["fw:version"] then 
		local fw_filename = fw_cfg["application"]
		local fw_dir  	  = "/var/www/firmwares/wildixpa/"
		local fw_file = fw_dir .. fw_filename
		if h.is_exists(fw_file) then
			logger:info("Found new firmware version - %s", fw_cfg["version"])
			local fw_link = string.gsub(
				"$protocol://$registrar:$port/$fwpath", "%$(%w+)",
				{
					registrar = registrar,
					protocol  = prov_proto,
					port 	  = prov_port,
					fwpath   = "firmwares/wildixpa/" .. fw_filename
				}
			)
			cfg_t["<AUTOUPDATE CONFIG MODULE>"]["Auto Image Url"] = fw_link
			logger:info("Firmware for device %s %s will be changed from version %s to %s", agent["model"], agent["mac"], agent["fw:version"], fw_cfg["version"])
		else
			logger:warning("Firmware file %q can't be loaded", fw_file)
		end
	end

----------------------------------------------------------------
--- General configuration settings
----------------------------------------------------------------
	cfg_t["<<VOIP CONFIG FILE>>Version"] = fw_cfg["config"] or "2.0001" -- confiugration version

---------------------------------------------------------
------- Access settings
---------------------------------------------------------
	-- WEB interface access settings
	cfg_t["<MMI CONFIG MODULE>"]["MMI Account"]["Account1 Name"] = "admin"
	cfg_t["<MMI CONFIG MODULE>"]["MMI Account"]["Account1 Password"] = token_pass

----------------------------------------------------------------
------- Assigned user data 
----------------------------------------------------------------
	-- get device setting from provision DB and username
	local sql = "SELECT a1.value, a2.value FROM autoprov AS a1 LEFT JOIN autoprov AS a2 ON (a1.mac = a2.mac AND a2.param=%q) WHERE a1.param=%q AND a1.mac=%q"
	local cursor = connection:execute(sql, "settings", "username", agent["mac"])
	local username, settings_json = cursor:fetch(); cursor:close()

	-- update device username information
	if not username then
		username = core.register(agent["mac"])
		if not username then
			logger:error("Can't set username to %s(%s). Provision is blocked", agent["model"] or "", agent["address"] or "")
			return {}, ""
		end
	end

	-- get LDAP data by username
	local extenNumber = h.get_extenNumber_str()
	local ldap_result = wmsldap:get(extenNumber, username)
	local person 	  = {}

	if not ldap_result then		-- somthing with LDAP?
		logger:error("Could not connect to LDAP server")
		return {}, username 
	elseif (0 == #ldap_result  or not ldap_result[1]) and not username:find("unknown%d") then	-- user not found in LDAP
		logger:debug("Could not get information for %q from LDAP", username)
		return {}, username
	elseif 0 ~= #ldap_result then
		_, person = unpack(ldap_result[1])
	end

	local secret = person["wuserPasswordSip"] or "1234"
	if username:find("unknown%d") then
		local rv, err = h.pass_for_unknown(username)
		if err then
			logger:error("password generation error for unknown account:", err)
		end
		secret = rv
	end

	if not person["preferredLanguage"] then
		person["preferredLanguage"] = h.get_admin_lang()
	end
	local user_lang  = person["preferredLanguage"] or "en"
	local user_group = person["owner"] or ""

	-- load user features
	local user_features = person.uid and require"common.userFeatures.userFeatures":load(person.uid) or {}

-------------------------------------------------------
----- Network configuration
-------------------------------------------------------	
	local net_cfg			 = {}
	local update_db_settings = false
	if settings_json then
		net_cfg = json.decode(settings_json)
	end
	local is_addr_remote 	 = not h.isLocalIP(agent["address"])

	-- SYSLOG server (by default syslog server is empty)
	if net_cfg["syslog_server"] and "" ~= net_cfg["syslog_server"] then
		cfg_t["<AAA CONFIG MODULE>"]["Enable Syslog"] = "1"
		cfg_t["<AAA CONFIG MODULE>"]["Syslog address"] = net_cfg["syslog_server"]
		cfg_t["<DEBUG CONFIG MODULE>"]["MGR Trace Level"] = "7"
		cfg_t["<DEBUG CONFIG MODULE>"]["SIP Trace Level"] = "7"
	end

	--TODO -- SNMP acess server

	-- VLAN settings
	if net_cfg["vid_voice"] and "" ~= net_cfg["vid_voice"] and 0 < tonumber(net_cfg["vid_voice"]) then
		cfg_t["<QOS CONFIG MODULE>"]["Enable VLAN"]         = "1"
		cfg_t["<QOS CONFIG MODULE>"]["VLAN ID"]             = net_cfg["vid_voice"] or "0"
		cfg_t["<QOS CONFIG MODULE>"]["Signalling Priority"] = net_cfg["qos_voice"] or "7"
		cfg_t["<QOS CONFIG MODULE>"]["Voice Priority"]      = net_cfg["qos_voice"] or "7"
	end
	if net_cfg["vid_data"] and "" ~= net_cfg["vid_data"] and 0 < tonumber(net_cfg["vid_data"]) then
		cfg_t["<QOS CONFIG MODULE>"]["Enable PVID"] = "2"
		cfg_t["<QOS CONFIG MODULE>"]["PVID Value"]  = net_cfg["vid_data"] or "0" 
	end

	-- set DNS and NTP settings by pbx network configuration and device setttings
   	net_cfg["dns_server"], net_cfg["second_dns_server"] = h.get_prim_sec_dns_servers(net_cfg["dns_server"], net_cfg["second_dns_server"])
	net_cfg["ntp_server"] = net_cfg["ntp_server"] or ""

	-- checking static option with device location
	local is_addr_present = h.is_addr_present(agent["address"], agent["mac"])	-- check dublicate address
	if net_cfg["use_static_ip"] and (is_addr_present or is_addr_remote) then
		if is_addr_present then
			logger:warning("Disable static ip. Other device have the same ip address: %s", agent["address"] or "")
		end
		if is_addr_remote then	-- device behind NAT can't user static ip
			logger:warning("Disable static ip. Device ip address (%s) is placed behind NAT", agent["address"] or "")
		end
		net_cfg["use_static_ip"] = nil
		update_db_settings 		 = true
	end

	-- get settings for leased device by pbx dhcp
	local is_leased	= false
	if not net_cfg["use_static_ip"] or "" == net_cfg["dns_server"] or "" == net_cfg["ntp_server"] then
		is_leased = h.is_addr_leased(agent["address"])
	end
	if not net_cfg["use_static_ip"] and not is_addr_present and is_leased then
		h.set_addr_static(agent["address"], agent["mac"])	-- mark this lease as static
		net_cfg["use_static_ip"] = 1
		update_db_settings 		 = true
	end

	-- set settings for leased device
	if is_leased then
		if "" == net_cfg["dns_server"] or "" == net_cfg["second_dns_server"] then
		local dns_dhcp1, dns_dhcp2 = h.get_dns_servers()
			if dns_dhcp1 and "" ~= dns_dhcp1 and "" == net_cfg["dns_server"] then
				net_cfg["dns_server"] = dns_dhcp1			 -- set primary DNS from DHCP config
				update_db_settings = true
			end
			if dns_dhcp2 and "" ~= dns_dhcp2 and "" == net_cfg["second_dns_server"] then
				net_cfg["second_dns_server"] = dns_dhcp2	  -- set secondary DNS from DHCP config
				update_db_settings = true
			end
		end

		if "" == net_cfg["ntp_server"] then -- set NTP from DHCP config
			local ntp_dhcp = h.get_ntp_servers()
			if ntp_dhcp and "" ~= ntp_dhcp then
				net_cfg["ntp_server"] = ntp_dhcp
				update_db_settings	= true
			end
		end
	end

	-- set settings for device with static options
	if 1 == net_cfg["use_static_ip"] then	-- use static ip settings
		local _, netmask, gateway = h.host_settings(agent["address"], registrar_ip)
		if netmask and gateway then	
			cfg_t["<GLOBAL CONFIG MODULE>"]["WAN Mode"] ="STATIC"
			cfg_t["<GLOBAL CONFIG MODULE>"]["Enable DHCP"] = "0"
			cfg_t["<GLOBAL CONFIG MODULE>"]["WAN IP"] = agent["address"]
			cfg_t["<GLOBAL CONFIG MODULE>"]["WAN Subnet Mask"] = netmask
			cfg_t["<GLOBAL CONFIG MODULE>"]["WAN Gateway"] = gateway
			
			if registrar_ip then
				if "" == net_cfg["dns_server"] then
					net_cfg["dns_server"] = registrar_ip
					update_db_settings 	  = true
				end
				if "" == net_cfg["second_dns_server"] then
					net_cfg["second_dns_server"] = registrar_ip
					update_db_settings 			 = true
				end
			end
	
			if "" == net_cfg["ntp_server"] then
				net_cfg["ntp_server"] = registrar
				update_db_settings	= true
			end
		else
			logger:warning("For current network configurations is not possible set static ip, disable it")
			net_cfg["use_static_ip"] = nil
			update_db_settings	   = true
		end
	end

	-- update devdb with new settings
	if update_db_settings then
		sql = "INSERT OR REPLACE INTO autoprov (id, param, value, mac) VALUES ((SELECT id FROM autoprov WHERE param=%q AND mac=%q), %q, '%s', %q)"
		connection:execute (sql, "settings", agent["mac"], "settings", json.encode(net_cfg), agent["mac"])
	end

	-- DNS settings 
	if "" ~= net_cfg["dns_server"] then
		cfg_t["<GLOBAL CONFIG MODULE>"]["DHCP Auto DNS"] = "0"
		cfg_t["<GLOBAL CONFIG MODULE>"]["Primary DNS"]	 	  = net_cfg["dns_server"]
		if "" ~= net_cfg["second_dns_server"] then
			cfg_t["<GLOBAL CONFIG MODULE>"]["Secondary DNS"] 	  = net_cfg["second_dns_server"]
		end
	end

	-- NTP settings
	cfg_t["<GLOBAL CONFIG MODULE>"]["Enable SNTP"] = "1"
	cfg_t["<GLOBAL CONFIG MODULE>"]["SNTP Server"] = net_cfg["ntp_server"] or "pool.ntp.org"

	-- Time zone
	local timeZone, location = __getTimeZoneLocation(net_cfg["ntp_timezone"] or h.getPbxTimeZone())
	cfg_t["<GLOBAL CONFIG MODULE>"]["Time Zone"] = timeZone or "4"
	cfg_t["<GLOBAL CONFIG MODULE>"]["Location"] = location or "13"

	-- Daylight saving time
	if "no" == net_cfg["use_dst"] then
		cfg_t["<GLOBAL CONFIG MODULE>"]["Enable DST"] = "0"
	end

	-- ToS, tonezone settings
	local tos_data, toneZone = h.get_sip_settings()
	

	-- Jitter buffer settings
	local tjb = h.get_jitter_buffer(is_addr_remote)

	-- Multicast paging
	if connection:switch("wmsdb") then
		local sql	  = "SELECT p.id, p.name FROM paging AS p LEFT JOIN paging_members AS pm ON (pm.paging_id=p.id) WHERE pm.group_member=%q"
		local res_sql = connection:rows(sql, username)
		if res_sql then
			local i = 1
			for id, name in res_sql do	-- filling value from table paging
				cfg_t["<MCAST CFG MODULE>"]["Mcast Addr"]["MCAST"..i.." Name"] = name
				cfg_t["<MCAST CFG MODULE>"]["Mcast Addr"]["MCAST"..i.." Host"] = "239.1.1.10"
				cfg_t["<MCAST CFG MODULE>"]["Mcast Addr"]["MCAST"..i.." Port"] = "1000"..(id-1)
				i = i + 1
			end
		end
	end

---------------------------------------------------------
------- User features settings
---------------------------------------------------------
		--TODO -- Call waiting tone feature

---------------------------------------------------------
------- Account settings
---------------------------------------------------------
	-- retrieve failover configuration
	local pbxFailover	 = nil
	local failoverSipPort = nil
	if not username:find("unknown%d") then
		local failoverObj = require"common.failover":new()
		if failoverObj and "1" == failoverObj:isEnabled() then
			local failoverCfg = failoverObj:getConfig()
			pbxFailover = failoverCfg and failoverCfg.pbxFailoverDomain
			failoverSipPort = failoverCfg and failoverCfg.sipPort
			logger:debug(failoverCfg, "failoverCfg")
		end
	end

	-- set SIP transport and SIP port
	local gtransp	= nil
	local gsipport	= nil
	if is_addr_remote then
		gtransp  = gconf.SIPTransportRemote
		gsipport = gconf.SIPServerPortRemote
	else
		gtransp  = gconf.SIPTransport
		gsipport = gconf.SIPServerPort
	end

	local sip_transport = "TLS"
	if gtransp then
		sip_transport   = gtransp
	end

	local sip_port   = "5060"
	if gsipport then
		sip_port	 = gsipport
	elseif "TLS" == sip_transport then
		sip_port	 = "5061"
		if is_addr_remote then
			sip_port = agent["alternate_port"] or "443"
		end
	elseif "TCP" == sip_transport then
		sip_port	 = "5060"
	end

	if pbxFailover then
		sip_transport   = "TCP"
		sip_port = agent["alternate_port"] or "443"
	end

	-- set SIP server settings
	cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP1 Phone Number"]  = username
	cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP1 Display Name"]  = person["cn"] or username
	cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP1 Register User"] = username
	cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP1 Register Pswd"] = secret
	cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP1 Proxy User"]    = username
	cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP1 Proxy Pswd"]    = secret
	cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP1 Register Addr"] = registrar
	cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP1 Transport"]	= ("TCP" == sip_transport and "1") or ("TLS" == sip_transport and "3") or "0"
	cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP1 Register Port"] = sip_port
	cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP1 Proxy Addr"]    = registrar
	cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP1 Proxy Port"]    = sip_port
	if pbxFailover then
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 Phone Number"]  = username
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 Display Name"]  = person["cn"] or username
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 Register User"] = username
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 Register Pswd"] = secret
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 Proxy User"]    = username
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 Proxy Pswd"]    = secret
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 Enable Reg"]    = "1"
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 Register Addr"] = pbxFailover
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 Transport"]	= ("TCP" == sip_transport and "1") or ("TLS" == sip_transport and "3") or "0"
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 Register Port"] = failoverSipPort
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 Proxy Addr"]    = pbxFailover
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 Proxy Port"]    = failoverSipPort
	end

	-- audio codecs priority
	if not h.is_wideband(agent["address"]) then
		cfg_t["<DSP CONFIG MODULE>"]["Port Config"]["P1 Voice Codec1"] = "G729/8000"
		cfg_t["<DSP CONFIG MODULE>"]["Port Config"]["P1 Voice Codec2"] = "PCMU/8000"
		cfg_t["<DSP CONFIG MODULE>"]["Port Config"]["P1 Voice Codec3"] = "PCMA/8000"
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP1 VoiceCodecMap"] = "G729,G711U,G711A"
		cfg_t["<SIP CONFIG MODULE>"]["SIP Line List"]["SIP2 VoiceCodecMap"] = "G729,G711U,G711A"
	end

	--TODO -- set SRPT settings to optional

	--TODO -- Language settings

---------------------------------------------------------
------- Function key settings
---------------------------------------------------------
	local userKeys = core.fkeys_get(user_features, username)
	if userKeys and userKeys.stdKeys then
		for i = 1,2 do
			local key, value = string.match(userKeys.stdKeys[i] or "", "(%w+)%s+(%S+)")
			if ("SD" == key or "EXT" == key) and value then
				cfg_t["<PHONE CONFIG MODULE>"]["Function Key"]["Fkey"..i.." Type"] = "1"
				cfg_t["<PHONE CONFIG MODULE>"]["Function Key"]["Fkey"..i.." Value"] = value.."@1/f"
			end
		end
	end

-----------------------------------------------------
--- Result serialization
-----------------------------------------------------
	local rv = {}
	
	local function __extractSubsection(section, cfg, stag, etag)
		if "string" == type(cfg) or "number" == type(cfg) then
			rv[#rv+1] = section..string.rep(" ", paramMaxLen-string.len(section))..":"..cfg
		elseif "table" == type(cfg) then
			rv[#rv+1] = (stag or "")..section..(etag or "")
			for k,v in pairs(cfg) do
				__extractSubsection(k, v, "--", "--"..string.rep(" ", paramMaxLen-4-string.len(k))..":")
			end
		end 
	end
	
	for _,section in ipairs(resource.sections) do
		__extractSubsection(section, cfg_t[section])
		rv[#rv+1] = ""
	end

	rv[#rv+1] = "<<END OF FILE>>"

	return rv
end
function setuser (...) end
function setfeatures (...) end
